package flowserver

import (
	"encoding/json"
	"flow"
	"flow/registry"
	"log"
	"reflect"
	"strings"
	"time"
)

// Node that will contain registry src
type Node struct {
	ID            string            `json:"id"`
	Src           string            `json:"src"`
	Label         string            `json:"label"`
	DefaultInputs map[int]string    `json:"defaultInputs"`
	Prop          map[string]string `json:"prop"`
}

// Link that joins two nodes
type Link struct {
	From string `json:"from"`
	To   string `json:"to"`
	In   int    `json:"in"`
}

// Trigger that join two nodes on state change
type Trigger struct {
	From string   `json:"from"`
	To   string   `json:"to"`
	On   []string `json:"on"`
}

// FlowDocument flow document
type FlowDocument struct {
	Nodes    []Node    `json:"nodes"`
	Links    []Link    `json:"links"`
	Triggers []Trigger `json:"triggers"`
}

// FlowBuild build a flowGraph from incoming web ui json
func FlowBuild(rawData []byte, r *registry.R) (*flow.Flow, error) {
	doc := FlowDocument{[]Node{}, []Link{}, []Trigger{}}
	err := json.Unmarshal(rawData, &doc)
	if err != nil {
		return nil, err
	}

	f := flow.New()
	f.SetRegistry(r)
	nodeMap := map[string]Node{}
	for _, n := range doc.Nodes {
		nodeMap[n.ID] = n
	}
	inputMap := map[string]flow.Operation{}

	ninput := 0
	for _, n := range doc.Nodes {
		if n.Src == "Input" || n.Src == "Variable" || n.Src == "Const" {
			continue
		}

		// Find link refered as To
		entry, err := r.Entry(n.Src)
		if err != nil {
			return nil, err
		}

		param := make([]flow.Data, len(entry.Inputs))

		// Default inputs
		for i := range param {
			if v, ok := n.DefaultInputs[i]; ok {
				param[i] = v
			}
		}

		// Find links
		for _, l := range doc.Links {
			if l.To != n.ID {
				continue
			}
			// Define operators here
			from := nodeMap[l.From]
			switch from.Src {
			case "Input":
				inOp, ok := inputMap[n.ID]
				if !ok {
					inOp = f.In(ninput)
					inputMap[n.ID] = inOp
					ninput++
				}
				param[l.In] = inOp // By id perhaps
			case "Variable":
				param[l.In] = f.Var(from.ID, from.Prop["init"])
			case "Const":
				// XXX: Automate this in a func
				raw := from.Label
				//raw := from.Prop["value"]
				raw = strings.TrimSpace(raw)
				if raw[0] == '"' { // its a string
					log.Println("Unmashal string")
					var val string
					json.Unmarshal([]byte(raw), &val)
					param[l.In] = val
					continue
				}
				log.Println("Will unmarshal to input:", raw)
				newVal := reflect.New(entry.Inputs[l.In])
				err := json.Unmarshal([]byte(raw), newVal.Interface())
				if err != nil {
					// ignore error?
					log.Println("unmarshalling Error", err)
					//param[l.In] = nil
					//continue
				}
				param[l.In], _ = f.Const(newVal.Elem().Interface())
			default:
				param[l.In] = f.Res(from.ID)
			}
		}
		f.DefOp(n.ID, n.Src, param...)
	}
	//Trigger testing
	for _, t := range doc.Triggers {
		log.Println("Trigger:", t)
		log.Println("Attaching an hook per trigger test")

		f.Hook(flow.Hook{
			Any: func(name string, ID string, triggerTime time.Time, extra ...interface{}) {
				if name == "waiting" {
					return
				}
				if ID != t.From {
					return
				}
				exec := false
				for _, o := range t.On {
					if name == o {
						exec = true
						break
					}
				}
				log.Println("Should trigger on:", name, t.On)
				if !exec {
					log.Println("Mismatching trigger, but its a test")
				}

				log.Println("Grabbing an executor:", t.To)
				op := f.Res(t.To)
				log.Println("Processing this operator")
				go op.Process() // Background
				log.Println("Returning from operator")
			},
		})

	}

	return f, nil
}
